import { __rest } from "tslib";
import { request as httpRequest, } from "http";
import { request as httpsRequest } from "https";
import { URL } from "url";
import { identity as identityEnumerable, lift, } from "@reactive-js/async-enumerable";
import { createDisposableValue, } from "@reactive-js/disposable";
import { createRedirectHttpRequest, parseHttpResponseFromHeaders, writeHttpRequestHeaders, } from "@reactive-js/http";
import { createWritableAsyncEnumerator, transform, emptyReadableAsyncEnumerable, createReadableAsyncEnumerable, } from "@reactive-js/node";
import { createObservable, ofValue, concatMap, onNotify, scan, } from "@reactive-js/observable";
import { pipe, compose } from "@reactive-js/pipe";
import { supportedEncodings, createEncodingCompressTransform, getFirstSupportedEncoding, } from "./httpContentEncoding";
import { contentIsCompressible } from "./httpContent";
import { decodeHttpContentResponse } from "./httpResponse";
const spyScanner = ([uploaded, total], ev) => ev.type === 1
    ? [ev.chunk.length, total + uploaded]
    : [-1, total + uploaded];
const createOnSubscribe = (request, send, encodeContent, decodeHttpContentResponse) => (subscriber) => {
    var _a;
    subscriber.dispatch({ type: 1, request });
    const req = send();
    const onError = (cause) => {
        subscriber.dispose({ cause });
    };
    req.on("error", onError);
    const onResponse = (msg) => {
        subscriber.add(() => msg.destroy());
        const response = pipe(parseHttpResponseFromHeaders(msg.statusCode || -1, msg.headers, createReadableAsyncEnumerable(() => msg)), decodeHttpContentResponse);
        const disposableResponse = createDisposableValue(response, _ => {
            msg.destroy();
        });
        subscriber.dispatch({
            type: 4,
            request,
            response: disposableResponse,
        });
    };
    req.on("response", onResponse);
    const reqBodyEnumerator = createWritableAsyncEnumerator(req, subscriber);
    const doOnNotify = ([count, total]) => {
        const ev = count < 0
            ? { type: 3, request }
            : { type: 2, request, total };
        subscriber.dispatch(ev);
    };
    const spyEnumerator = pipe(identityEnumerable(), lift(compose(scan(spyScanner, () => [0, 0]), onNotify(doOnNotify)))).enumerateAsync(subscriber);
    const contentEnumerator = pipe(((_a = request.content) === null || _a === void 0 ? void 0 : _a.body) || emptyReadableAsyncEnumerable, lift(onNotify(ev => spyEnumerator.dispatch(ev))), encodeContent).enumerateAsync(subscriber);
    subscriber
        .add(reqBodyEnumerator)
        .add(spyEnumerator)
        .add(contentEnumerator)
        .add(_ => {
        req.abort();
        req.removeAllListeners();
        req.destroy();
    });
    const onContinue = () => {
        contentEnumerator.subscribe(reqBodyEnumerator);
        reqBodyEnumerator.subscribe(contentEnumerator);
    };
    if (request.expectContinue) {
        req.on("continue", onContinue);
    }
    else {
        onContinue();
    }
};
const identity = (x) => x;
const requestIsCompressible = (request) => {
    const { content } = request;
    return content !== undefined ? contentIsCompressible(content) : false;
};
export const creatHttpClient = (clientOptions = {}) => {
    const { agent, insecureHTTPParser, maxHeaderSize, shouldEncode: shouldEncodeOption } = clientOptions, zlibOptions = __rest(clientOptions, ["agent", "insecureHTTPParser", "maxHeaderSize", "shouldEncode"]);
    return (request, requestOptions = {}) => {
        const { acceptedEncodings = supportedEncodings } = requestOptions;
        const contentEncoding = getFirstSupportedEncoding(acceptedEncodings);
        const send = () => {
            const { content } = request;
            const newContent = content === undefined || contentEncoding === undefined
                ? content
                : Object.assign(Object.assign({}, content), { contentEncodings: [...content.contentEncodings, contentEncoding] });
            const requestWithAcceptEncodingsAndContentEncoding = Object.assign(Object.assign({}, request), { acceptedEncodings: supportedEncodings, content: newContent });
            const { method, uri } = requestWithAcceptEncodingsAndContentEncoding;
            const headers = {};
            writeHttpRequestHeaders(requestWithAcceptEncodingsAndContentEncoding, (header, value) => (headers[header] = value));
            const nodeRequestUrl = uri instanceof URL ? uri : new URL(uri.toString());
            const nodeRequestOptions = {
                agent,
                headers,
                insecureHTTPParser,
                maxHeaderSize,
                method,
            };
            return uri.protocol === "https:"
                ? httpsRequest(nodeRequestUrl, nodeRequestOptions)
                : uri.protocol === "http:"
                    ? httpRequest(nodeRequestUrl, nodeRequestOptions)
                    : (() => {
                        throw new Error();
                    })();
        };
        const shouldEncodeOptionResult = shouldEncodeOption !== undefined
            ? shouldEncodeOption(request)
            : undefined;
        const shouldEncode = shouldEncodeOptionResult !== undefined
            ? shouldEncodeOptionResult
            : requestIsCompressible(request);
        const contentEncoder = contentEncoding !== undefined && shouldEncode
            ? transform(createEncodingCompressTransform(contentEncoding, zlibOptions))
            : identity;
        const decodeHttpContentResponseWithOption = decodeHttpContentResponse(zlibOptions);
        return createObservable(createOnSubscribe(request, send, contentEncoder, decodeHttpContentResponseWithOption));
    };
};
const redirectCodes = [
    301,
    302,
    303,
    307,
    308,
];
export const createDefaultHttpResponseHandler = (sendHttpRequest, maxRedirects = 10) => {
    const handleResponse = (status) => {
        var _a;
        if (status.type === 4) {
            const { request, response } = status;
            const { location, preferences, statusCode } = response.value;
            const acceptedEncodings = ((_a = preferences) === null || _a === void 0 ? void 0 : _a.acceptedEncodings) || [];
            const shouldRedirect = redirectCodes.includes(statusCode) &&
                location !== undefined &&
                maxRedirects > 0;
            const [newRequest, newAcceptedEncodings] = shouldRedirect
                ? [
                    createRedirectHttpRequest(response.value)(request),
                ]
                : statusCode === 417
                    ? [Object.assign(Object.assign({}, request), { expectContinue: false })]
                    : statusCode === 415 &&
                        acceptedEncodings.length > 0
                        ? [request, acceptedEncodings]
                        : [request];
            if (request !== newRequest || newAcceptedEncodings !== undefined) {
                response.dispose();
                return pipe(sendHttpRequest(newRequest, {
                    acceptedEncodings: newAcceptedEncodings,
                }), createDefaultHttpResponseHandler(sendHttpRequest, maxRedirects - 1));
            }
        }
        return ofValue(status);
    };
    return concatMap(handleResponse);
};
//# sourceMappingURL=httpClient.js.map